#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>

#include "pty.h"

static int slavetty = -1;

/* Some processes set the terminal back to echo after e.g. asking for a
   password.  So we set it to noecho on each write to the child process. */
void
setslavenoecho(void)
{
    struct termios tio;
    
    if (tcgetattr(slavetty, &tio) < 0) {
	perror("ule: tcgetattr on slave");
	exit(1);
    }

    if (!(tio.c_lflag & (ECHO | ECHOE | ECHOK | ECHONL))) {
	/* No need to reset it. */
	return;
    }

    tio.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
    
    if (tcsetattr(slavetty, TCSANOW, &tio) < 0) {
	perror("ule: tcsetattr on slave");
	exit(1);
    }
}

/* Set termios of slave pty to that of the controlling terminal of the
   parent, with desired modifications. */
static void
setslavetermios(int fd, struct termios *tio)
{
    tio->c_lflag &= ~ECHO;
    tio->c_lflag &= ~ECHOE;
    tio->c_lflag &= ~ECHOK;
    tio->c_lflag &= ~ECHONL;
    
    if (tcsetattr(fd, TCSANOW, tio) < 0) {
	perror("ule: tcsetattr on slave");
	exit(1);
    }    
}

/* Get termios from current controlling terminal. */
static void
gettermios(struct termios *tio)
{
    int fd;

    fd = open("/dev/tty", O_RDWR);
    if (fd < 0) {
	perror("ule: open(\"/dev/tty\")");
	exit(1);
    }

    if (tcgetattr(fd, tio) < 0) {
	perror("ule: tcgetattr");
	exit(1);
    }

    close(fd);
}

pid_t
execpty(int *masterfd,
	const char *file, char *const argv[])
{
    struct termios tio;
    char devname[32];
    pid_t pid;

    gettermios(&tio);

    *masterfd = getmaster(devname);

    pid = fork();
    if (pid < 0) {
	perror("ule: fork()");
	exit(1);
    }

    if (pid > 0) {
	slavetty = getslave(devname);
	return pid;
    }
    
    close(*masterfd);

    if (setsid() < 0) {
	perror("ule: setsid()");
	exit(1);
    }

    slavetty = getslave(devname);
    setslavetermios(slavetty, &tio);
    
    /* Writing errors from here down is pretty useless as they will probably
       end up nowhere. */
    dup2(slavetty, STDIN_FILENO);
    dup2(slavetty, STDOUT_FILENO);
    dup2(slavetty, STDERR_FILENO);
    if (slavetty != STDIN_FILENO &&
	slavetty != STDOUT_FILENO &&
	slavetty != STDERR_FILENO) {
	close(slavetty);
    }

    execvp(file, argv);
    exit(1);
}
