#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stropts.h>

int
getmaster(char *devname)
{
    const char *name;

    const int fd = open("/dev/ptmx", O_RDWR);
    if (fd < 0) {
	perror("ule: open(\"/dev/ptmx\")");
	exit(1);
    }

    if (grantpt(fd) < 0) {
	perror("ule: grantpt");
	exit(1);
    }

    if (unlockpt(fd) < 0) {
	perror("ule: unlockpt");
	exit(1);
    }

    name = ptsname(fd);
    if (name == NULL) {
	perror("ule: ptsname");
	exit(1);
    }

    strcpy(devname, name);
    return fd;
}

int
getslave(const char *devname)
{
    const int fd = open(devname, O_RDWR);
    if (fd < 0) {
	fprintf(stderr, "ule: open(\"%s\"): %s\n", devname, strerror(errno));
	exit(1);
    }

    if (ioctl(fd, I_PUSH, "ptem") < 0) {
	perror("ule: ioctl(I_PUSH, \"ptem\")");
	exit(1);
    }
    if (ioctl(fd, I_PUSH, "ldterm") < 0) {
	perror("ule: ioctl(I_PUSH, \"ldterm\")");
	exit(1);
    }
    if (ioctl(fd, I_PUSH, "ttcompat") < 0) {
	perror("ule: ioctl(I_PUSH, \"ttcompat\")");
	exit(1);
    }

    return fd;
}
