#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>

#include "rlstuff.h"
#include "execpty.h"

static pid_t ule_slavepid = 0;

static void
handlesignal(int signo)
{
    int status;

    switch (signo) {
    case SIGHUP:
    case SIGQUIT:
    case SIGTERM:
	kill(ule_slavepid, signo);
	exit(128 + signo);
	break;
    case SIGINT:
	kill(ule_slavepid, signo);
	break;
    case SIGCHLD:
	if (waitpid(ule_slavepid, &status, WNOHANG) == ule_slavepid) {
	    exit(0);
	}
	break;
    }
}

static void
installhandlers(void)
{
    struct sigaction sa;

    sa.sa_handler = handlesignal;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;

    sigaction(SIGHUP, &sa, NULL);
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGQUIT, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGCHLD, &sa, NULL);
}

static void
do_childatexit(void)
{
    kill(ule_slavepid, SIGHUP);
}

static void
usage(void)
{
    fprintf(stderr, 
	    "ule 0.1, the Universal Line Editor\n"
	    "Usage: ule [options] command [args ...]\n"
	    "Options:\n"
	    "-b type Highlight input as type (if terminal supports it)\n"
	    "        type = s(tandout), b(old), f(lash), r(everse), u(nderline)\n"
	    "-h      Print this message\n");
}

int 
main(int argc, char **argv)
{
    int option;
    char highlight = '0';
    sigset_t childmask;
    int ule_ptyfd;
    fd_set fixedfds;

    while ((option = getopt(argc, argv, "b:h")) != EOF) {
	switch (option) {
	  case 'b':
	    highlight = tolower((unsigned char)optarg[0]);
	    break;
	  case 'h':
	    usage();
	    exit(0);
	  case '?':
	    usage();
	    exit(1);
	}
    }

    if (argc - optind < 1) {
	usage();
	exit(1);
    }

    installhandlers();
    atexit(do_childatexit);

    /* Block SIGCHLD temporarily to avoid getting the SIGCHLD before the slave
       pid has been stored in ule_slavepid. */
    sigemptyset(&childmask);
    sigaddset(&childmask, SIGCHLD);
    sigprocmask(SIG_BLOCK, &childmask, NULL);
    ule_slavepid = execpty(&ule_ptyfd, argv[optind], argv + optind);
    sigprocmask(SIG_UNBLOCK, &childmask, NULL);

    FD_ZERO(&fixedfds);
    FD_SET(STDIN_FILENO, &fixedfds);
    FD_SET(ule_ptyfd, &fixedfds);

    initialize_readline(ule_ptyfd, highlight);

    for (;;) {
	fd_set readyfds = fixedfds;
	
	if (select(ule_ptyfd + 1, &readyfds, NULL, NULL, NULL) < 0) {
	    if (errno == EINTR) {
		continue;
	    }
	    perror("ule: select()");
	    exit(1);
	}
	if (FD_ISSET(STDIN_FILENO, &readyfds)) {
	    do_stdinput();
	}
	if (FD_ISSET(ule_ptyfd, &readyfds)) {
	    do_childinput();
	}
    }
}
